import multiprocessing
from multiprocessing.sharedctypes import Value as mpValue
import ctypes
import queue
import time
import psutil

from pyrateshield import labels, Logger, GLOBAL_LOG_LEVEL

from pyrateshield.radtracer import radtracer
from pyrateshield.pyshield.engine import Engine
import pandas as pd

LOG_LEVEL = GLOBAL_LOG_LEVEL


CRITICAL_POINT_NAME         = 'Critical Point Name'
RADTRACER_DOSE              = 'RadTracer Dose [mSv]'
PYSHIELD_DOSE               = 'PyShield Dose [mSv]'
OCCUPANCY_FACTOR            = 'Occupancy factor'
RADTRACER_DOSE_CORRECTED    = 'RadTracer Dose corrected for occupancy [mSv]'
PYSHIELD_DOSE_CORRECTED     = 'PyShield Dose corrected for occupancy [mSv]'
SOURCE_NAME                 = 'Source Name'

COLUMNS = [CRITICAL_POINT_NAME, PYSHIELD_DOSE, RADTRACER_DOSE, 
           OCCUPANCY_FACTOR, PYSHIELD_DOSE_CORRECTED, RADTRACER_DOSE_CORRECTED]

SPLIT_COLUMNS = [CRITICAL_POINT_NAME, SOURCE_NAME, PYSHIELD_DOSE, 
                 RADTRACER_DOSE, OCCUPANCY_FACTOR, 
                 RADTRACER_DOSE_CORRECTED, PYSHIELD_DOSE_CORRECTED]


class Worker(multiprocessing.Process, Logger):
    def __init__(self, project_queue, source_queue, dosemap_queue, update_flag, stop_flag):
        multiprocessing.Process.__init__(self)
        Logger.__init__(self, log_level=LOG_LEVEL)

        self.project_queue = project_queue
        self.source_queue = source_queue
        self.dosemap_queue = dosemap_queue
        self.update_flag = update_flag
        self.stop_flag = stop_flag        
    
    def update_project(self):        
        if self.project.dosemap.engine == labels.PYSHIELD:
            self.pyshield_engine = Engine.from_pyrateshield(self.project)
    
    def get_dosemap(self, source):
        if self.project.dosemap.engine == labels.PYSHIELD:
            return self.pyshield_engine.source_dosemap(source)
        elif self.project.dosemap.engine == labels.RADTRACER:
            return radtracer.dosemap_single_source(source, self.project)
            
    def run(self):
        while not self.stop_flag.value:            
            try:
                sources = self.source_queue.get(timeout=0.1)
            except queue.Empty:
                sources = None
            
            if self.update_flag.value:
                self.project = self.project_queue.get()
                self.update_project()
                self.update_flag.value = False
                self.logger.info(f'{self.name}, Updated project')
        
            if sources is not None:                
                for source in sources:                    
                    self.dosemap_queue.put( self.get_dosemap(source) )
                    
        self.logger.info(f'{self.name}, Stopped')


class Dosemapper(Logger):
    def __init__(self, processes=None, multi_cpu=True):
        Logger.__init__(self, log_level=LOG_LEVEL)
        
        self.multi_cpu = multi_cpu
        
        if self.multi_cpu:
            self.source_queue = multiprocessing.Queue()
            self.dosemap_queue = multiprocessing.Queue()
            self._workers = []
            
            self.max_cpu_count = 4 #psutil.cpu_count(logical=True)
            if processes is None:
                processes = psutil.cpu_count(logical=False)
            self.set_workers(processes)
    
    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._stop_workers()
    
    def update_project(self, project):
        # Make sure all worker processes are done updating from a previous call
        # to this method before updating again.
        while any( w.update_flag.value for w in self._workers ):
            time.sleep(0.1)
        
        for w in self._workers:
            w.project_queue.put(project)
            w.update_flag.value = True

    @staticmethod
    def get_single_cpu_dosemap(project, sources=None):
        # DEBUGGING Purpuses
        if sources is None:
            sources = list(project.sources_nm) + list(project.sources_ct)\
                + list(project.sources_xray)
                
        dosemap = None        
        update_dosemap = lambda dm: dm if dosemap is None else (dosemap + dm)
            
        if project.dosemap.engine == labels.PYSHIELD:
            pyshield_engine = Engine.from_pyrateshield(project)
        
        for source in sources:
            if not source.enabled:
                continue

            if source.label == labels.SOURCES_NM and project.dosemap.engine == labels.PYSHIELD:
                source_dosemap = pyshield_engine.source_dosemap(source)
            else:
                source_dosemap = radtracer.dosemap_single_source(source, project)
        
            dosemap = update_dosemap(source_dosemap)
        return dosemap


    def get_dosemap(self, project, sources=None):
        if not self.multi_cpu:
            return self.get_single_cpu_dosemap(project, sources)
        # Process all sources by default
        if sources is None:
            sources = list(project.sources_nm) + list(project.sources_ct) + list(project.sources_xray)
        
        # Exclude sources which are not set to Enabled
        sources = [src for src in sources if src.enabled]
        
        
        # Check if there is any work to be done
        if not len(sources):
            return None
        
                
        #This is a bit of a hack to ensure PyShield only receives NM sources,
        #and that sources with the same position are processed by the same worker.
        #This allows PyShield to benefit from caching.
        sources_non_nm = [src for src in sources if src.label != labels.SOURCES_NM]
        if project.dosemap.engine == labels.PYSHIELD:
            src_pos_dct = {}
            sources_nm = [src for src in sources if src not in sources_non_nm]

            for source in sources_nm:
                # Make a dictionary of source_positions and group the corresponding sources
                # I.e.: { (x,y,z): [src1, src2, src3 ], (x,y,z): [src4, src5, ... ] }
                src_pos_dct.setdefault(tuple(source.position), []).append(source)
            sources_grouped = [srcs for srcs in src_pos_dct.values()]
        else:
            sources_grouped = [[src] for src in sources]
        
                       
        self.update_project(project)
        
        for grp in sources_grouped:
            self.source_queue.put(grp)
        
        # Collect the results. Make sure to get as many results from the queue
        # as the number of sources
        nr_of_sources = sum(len(grp) for grp in sources_grouped)
                
        dosemap = None        
        update_dosemap = lambda dm: dm if dosemap is None else (dosemap + dm)
        
        for i in range(nr_of_sources):
            dosemap = update_dosemap( self.dosemap_queue.get() )        
        
        if project.dosemap.engine == labels.PYSHIELD and len(sources_non_nm):
            # Temporarily switch to RadTracer to process the non-NM sources
            print(f"Switching to RadTracer for {len(sources_non_nm)} non-NM sources")
            project.dosemap.engine = labels.RADTRACER
            dosemap = update_dosemap( self.get_dosemap(project, sources=sources_non_nm) )
            print("Done... Switching back to pyShield")
            project.dosemap.engine = labels.PYSHIELD
        
        return dosemap        
    
    
    @staticmethod
    def empty_critical_point_result():
        return pd.DataFrame([{CRITICAL_POINT_NAME:      None,
                              SOURCE_NAME:              None,
                              RADTRACER_DOSE:           None,
                              PYSHIELD_DOSE:            None,
                              OCCUPANCY_FACTOR:         None,
                              RADTRACER_DOSE_CORRECTED: None,
                              PYSHIELD_DOSE_CORRECTED:  None
                              }])
    @staticmethod
    def get_source_critical_point_result(source, 
                                         critical_point, 
                                         dose_radtracer, 
                                         dose_pyshield):
        
        factor = critical_point.occupancy_factor
        
        return {CRITICAL_POINT_NAME:        critical_point.name,
                SOURCE_NAME:                source.name,
                RADTRACER_DOSE:             dose_radtracer,
                PYSHIELD_DOSE:              dose_pyshield,
                OCCUPANCY_FACTOR:           factor,
                RADTRACER_DOSE_CORRECTED:   factor * dose_radtracer,
                PYSHIELD_DOSE_CORRECTED:    factor * dose_pyshield}
                              
        
    @staticmethod
    def get_sources_critical_points(project):
        sources = list(project.sources_nm) + list(project.sources_ct) + list(project.sources_xray)
        
        # Exclude sources which are not set to Enabled
        sources = [src for src in sources if src.enabled]
        
        # Exclude points which are not set to Enabled
        crit_points = [crp for crp in project.critical_points if crp.enabled]
                
        if len(crit_points) == 0 or len(sources) == 0:            
            return Dosemapper.empty_critical_point_result()
        
        reports = []        
        
        pyshield_engine = Engine.from_pyrateshield(project)
        
        for crit_point in crit_points:            
            for source in sources:
                dose_radtracer = radtracer.pointdose_single_source(
                    crit_point.position, source, project)

                if source.label == labels.SOURCES_NM:
                    dose_pyshield = pyshield_engine.dose_at_point(crit_point.position,
                                                         sources=[source])
                else:
                    dose_pyshield = dose_radtracer

                reports += [Dosemapper.get_source_critical_point_result(
                    source, crit_point, dose_radtracer, dose_pyshield)]
        report = pd.DataFrame(reports)[SPLIT_COLUMNS]
        
        return report
        
    
    def get_critical_points(self, project):
        
        
        if len(project.sources_nm) + len(project.sources_ct) + len(project.sources_xray) == 0\
            or len(project.critical_points) == 0:
                return self.empty_critical_point_result()
            
        report = self.get_sources_critical_points(project)
        
        aggfunc = {RADTRACER_DOSE: sum, RADTRACER_DOSE_CORRECTED: sum,
                   PYSHIELD_DOSE: sum, PYSHIELD_DOSE_CORRECTED: sum,
                   OCCUPANCY_FACTOR: max}
        
        summed_report = pd.pivot_table(report, index=CRITICAL_POINT_NAME, 
                                       aggfunc=aggfunc)
        
        # preserve order
        pnames = [crit_point.name for crit_point in project.critical_points\
                  if crit_point.enabled]
            
        summed_report = summed_report.reindex(pnames)
        
        summed_report = summed_report.reset_index()
       
        summed_report = summed_report[COLUMNS]
        return summed_report

    @property
    def n_workers(self):
        return len(self._workers)
    
    
    def set_workers(self, processes):
        if processes > self.max_cpu_count:
            self.logger.info(f"Max nr of parallel processes: {self.max_cpu_count}")
            processes = self.max_cpu_count
        if processes < 0:
            processes = 0
        
        
        if processes > self.n_workers:
            self._start_workers(processes - self.n_workers)
        elif processes < self.n_workers:
            self._stop_workers(self.n_workers - processes)
    
    
    def _start_workers(self, processes=None):
        self.logger.info(f"Starting {processes} workers")
        
        for i in range(processes):
            project_queue = multiprocessing.Queue()
            stop_flag = mpValue(ctypes.c_bool, False)
            update_flag = mpValue(ctypes.c_bool, False)
            args = (
                project_queue,
                self.source_queue,
                self.dosemap_queue,
                update_flag,
                stop_flag,
            )
            p = Worker(*args)            
            self._workers.append(p)
            p.start()
        
        
    def _stop_workers(self, processes=None):
        if processes is None:
            processes = self.n_workers
        
        stopworkers = [self._workers.pop() for i in range(processes)]
                    
        for w in stopworkers:
            w.stop_flag.value = True
        for w in stopworkers: 
            w.join()
        time.sleep(0.1)
        for w in stopworkers:
            if w.is_alive():
                self.logger.info("Stil alive")
            w.close()
        
            
if __name__ == "__main__":
    from pyrateshield.model import Model
    import timeit
    import time
    #model = Model.load_from_project_file('/Users/marcel/git/pyrateshield/example_projects/LargeProject/large_project.psp')
    #model = Model.load_from_project_file('../example_projects/LargeProject/large_project.psp')
    model = Model.load_from_project_file('../example_projects/SmallProject/project.psp')
    #model = Model.load_from_project_file('../example_projects/Lu-177.psp')
    model.dosemap.grid_matrix_size = 120
    
    with Dosemapper() as dm:        
        model.dosemap.engine = labels.PYSHIELD
        print("PyShield", timeit.timeit(lambda: dm.get_dosemap(model), number=1) )
    
        time.sleep(1) 
        print()
        
        model.dosemap.engine = labels.RADTRACER
        print("Radtracer", timeit.timeit(lambda: dm.get_dosemap(model), number=1) )

    # from pyshield import Sources
    # sources = Sources.from_pyrateshield(model)
    # print("Pyshield single CPU:", end=" ", flush=True)
    # print(timeit.timeit(lambda: sources.get_dosemap(), number=1) )
    
    
    
    





from pyrateshield import labels, __version__
from pyrateshield.gui.io import confirm_changes
from pyrateshield.gui.item_views import (EditPixelSizeView, 
                                         EditShieldingView, 
                                         EditCriticalPointsView,  
                                         EditSourcesNMView,
                                         EditWallsView,
                                         EditSourceCTView, 
                                         EditSourceXrayView,
                                         EditMaterialsView,
                                         EditClearanceView)
                                         
                                      
                                      


from PyQt5.QtWidgets import (QWidget, QLabel, QToolBox, QTabWidget, QMainWindow, 
                             QSplitter, QApplication, QPushButton, QTableView,
                             QHBoxLayout, QVBoxLayout, QGridLayout, QCheckBox)

from pyrateshield.gui.mpl_view import NavigationToolbar, MplCanvas



TOOLBOX_LAYOUT = {'Nuclear Medicine Sources':       [labels.SOURCES_NM, 
                                                     labels.CLEARANCE],
                  'Radiology Sources':              [labels.SOURCES_CT,
                                                     labels.SOURCES_XRAY],
                                    
                  'Walls, Shieldings && Materials':   [labels.WALLS,
                                                      labels.SHIELDINGS,
                                                      labels.MATERIALS],
                  labels.CRITICAL_POINTS:           [labels.CRITICAL_POINTS],
                  labels.PIXEL_SIZE_CM:             [labels.PIXEL_SIZE_CM]}                                                
                                              
                

RESULT_LAYOUT = [labels.CANVAS, labels.CRITICAL_POINT_REPORT_VIEW]

        
class ResultDosemapView(QWidget):
    dosemap_button_pyshield_label   = "View Dose Map PyShield"
    dosemap_button_radtracer_label   = "View Dose Map RadTracer"
    
    
    refresh_button_label    = "Refresh"
    _buttons = None
    
    @property
    def ncolumns(self):
        return len(self.buttons)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.create_widgets()
        self.create_layout()
        
    def create_widgets(self):

        self.canvas = MplCanvas()
        self.toolbar = NavigationToolbar(self.canvas, self)        
        self._labels = [
            self.dosemap_button_pyshield_label,
            self.dosemap_button_radtracer_label,
            self.refresh_button_label
        ]        
        self.buttons = {label: QPushButton(label) for label in self._labels}

    def create_layout(self):
        btn_layout = QHBoxLayout()
        for label in self._labels:
            btn_layout.addWidget(self.buttons[label])        
        
        #btn_layout.addWidget(self.toolbar)
        
        layout = QVBoxLayout(self)
        layout.addWidget(self.toolbar)
        layout.addWidget(self.canvas)
        #layout.addLayout(btn_layout)

        
        
class ResultCriticalPointView(QWidget):
    check_button_label = 'Show contribution of each individual source'
    def __init__(self, parent=None):
        super().__init__(parent)
        self.create_widgets()
        self.create_layout()
        
    def create_widgets(self):
        self.critical_point_button        = QPushButton("Calculate Critical Points")
        self.save_critcial_point_button   = QPushButton("Save To Excel")
        
        self.table_view             = QTableView()
        self.table_view.verticalHeader().setVisible(False)
        self.source_checkbox = QCheckBox(self.check_button_label)

        #self.table_view.setRowCount(100)
        #self.table_view.setColumnCount(5)
        
    def create_layout(self):
        vlayout = QVBoxLayout()

        vlayout.addWidget(self.table_view)
    
        
        blayout = QGridLayout()
        blayout.addWidget(self.critical_point_button, 0, 0)
        blayout.addWidget(self.save_critcial_point_button, 0, 1)
        
        vlayout.addLayout(blayout)
        vlayout.addWidget(self.source_checkbox)
        
        self.setLayout(vlayout)
        
        # auto resize columns
        header = self.table_view.horizontalHeader()
        header.setSectionResizeMode(header.ResizeToContents)
        
        
    
        
class MainView(QMainWindow):
    _views = None
    def __init__(self, mouse=None):
        super(MainView, self).__init__()
        self.views = self.create_widgets()
        self.create_layout()
        self.status_label = self.create_status_label()
        self.show()
        self.set_focus(labels.SOURCES_NM)
        
        
        
        
        self.setWindowTitle('PyrateShield (version ' + __version__ + ")")
    
    def closeEvent(self, event):
        if confirm_changes('Quit Application?'):
            super().closeEvent(event)
        else:
            event.ignore()
    
    
    def get_active_tool_panel_name(self):
        tool_index = self.toolbox.currentIndex()
        tool_name = self.toolbox.itemText(tool_index)
        if len(TOOLBOX_LAYOUT[tool_name]) > 1:
            tab_index = self.toolbox_tabs[tool_name].currentIndex()
        else:
            tab_index = 0
            
        tab_name = TOOLBOX_LAYOUT[tool_name][tab_index]
        
        return tab_name
    
    
    def set_focus(self, item_label):
        def location_in_toolbox(label):
            for name, items in TOOLBOX_LAYOUT.items():
                if label in items:
                    return name
            return None
            
        
        if item_label in RESULT_LAYOUT:
            self.result_container.setCurrentWidget(self.views[item_label])
        
        toolgroup = location_in_toolbox(item_label)
        
        if toolgroup is None:
            raise ValueError(f'No view item with label {item_label}')
            
        else:
            toolbox_tab = self.toolbox_tabs[toolgroup]
            self.toolbox.setCurrentWidget(toolbox_tab)
            if len(TOOLBOX_LAYOUT[toolgroup]) > 1:
                toolbox_tab.setCurrentWidget(self.views[item_label])
            
    def set_tab_enabled(self, label, enabled):
        for toolbox_tab_name, tabs in TOOLBOX_LAYOUT.items():
            if label in tabs:
                index = tabs.index(label)
                tab_widget = self.toolbox_tabs[toolbox_tab_name]
                tab_widget.setTabEnabled(index, enabled)
                
            
    @property
    def status_text(self):
        return self.status_label.text()
    
    def set_status_text(self, text):
        self.status_text = text
    
    @status_text.setter
    def status_text(self, text):
        self.status_label.setText(str(text))
        
        
    
    def create_widgets(self):
        if self._views is None:
            views = {labels.SOURCES_NM:             EditSourcesNMView(),
                     labels.CLEARANCE:              EditClearanceView(),
                     labels.SOURCES_CT:             EditSourceCTView(),
                     labels.SOURCES_XRAY:           EditSourceXrayView(),
                     # labels.FLOORPLAN:              EditFloorplanView(),
                     labels.PIXEL_SIZE_CM:          EditPixelSizeView(),
                   
                     labels.CRITICAL_POINTS:        EditCriticalPointsView(),
                     labels.WALLS:                  EditWallsView(),
                     labels.SHIELDINGS:             EditShieldingView(),
                     labels.MATERIALS:              EditMaterialsView(),
                     
                     labels.CANVAS:                 ResultDosemapView(),
                     labels.CRITICAL_POINT_REPORT_VIEW:  ResultCriticalPointView()}
            self._views = views
                 
        
        
        return self._views
                 
                  
                 
        
    def create_status_label(self):
        status_label = QLabel()
        
        statusbar = self.statusBar()
        statusbar.addWidget(status_label)
        statusbar.setVisible(True)
        return status_label
        

        

    def put_views_in_tabs(self, layout):
        tab_widget = QTabWidget()
        for item in layout:
            tab_widget.addTab(self.views[item], item)
        return tab_widget
        
    def put_views_in_toolbox(self, layout):
        toolbox = QToolBox()
        tabs = {}
        for group_name, items in layout.items():
            if len(items) > 1:
                tabs[group_name] = self.put_views_in_tabs(items)
            else:
                tabs[group_name] = self.views[items[0]]
            toolbox.addItem(tabs[group_name], group_name)
        
        return toolbox, tabs
       
    
    def create_layout(self):
        toolbox, toolbox_tabs = self.put_views_in_toolbox(TOOLBOX_LAYOUT)
        self.toolbox = toolbox
        self.toolbox_tabs = toolbox_tabs
        
        self.result_container = self.put_views_in_tabs(RESULT_LAYOUT)
        
        self.main_container = QSplitter()
        
        # self.toolbox.setMinimumWidth(WIDGET_WIDTH+10)
        #self.toolbox.setMaximumWidth(WIDGET_WIDTH+20)
        
        
        self.main_container.addWidget(self.toolbox)
       
        self.main_container.addWidget(self.result_container)
        
        

        self.setCentralWidget(self.main_container)
        
        self.setContentsMargins(10, 10, 10, 10)
        
        self.main_container.setSizes([100, 500])

    
        
if __name__ == "__main__":
    def main():
        app = QApplication([])
        window = MainView()
        window.show()    
        app.exec_()
        return window
    
    
    window = main()

        



import yaml
import os




_file = os.path.join(os.path.split(__file__)[0],
                     'styles.yml')

STYLES = yaml.safe_load(open(_file))

SELECTED = 'selected'
NORMAL = 'normal'
DEFAULT = 'default'
DISABLED = 'disabled'



# -*- coding: utf-8 -*-
"""
Created on Mon Feb 21 14:41:18 2022

@author: 757021
"""
import yaml
from collections.abc import MutableSequence
from pyrateshield import labels, Observable, GLOBAL_LOG_LEVEL
from pyrateshield.model_items import Clearance


LOG_LEVEL = GLOBAL_LOG_LEVEL

class ModelItemSequence(MutableSequence, Observable):
    # Sequence for Walls
    #
    # Sequence emulates a list. Items can be appended, inserted or set.
    # 
    # Sequences contain classes derived from ModelItem
    # Each model item can trigger an 'event_update'. THese events are 
    # catched by the sequence and resend as an 'event_update_item' by the
    # sequence. THe GUI only needs to connect to the sequence to receive all
    # events form each ModelItem.
    #
    # In Addition 'event_remove_item' will be emmitted when a ModelItem is 
    # removed. 'event_will_remove_item' just prior to removal.
    #'event_add_item' is emmitted when a new model item is added to the 
    # sequence.

    
    EVENT_REMOVE_ITEM       = 'event_remove_item'
    EVENT_WILL_REMOVE_ITEM  = 'event_will_remove_item' 
    EVENT_ADD_ITEM          = 'event_add_item'
    EVENT_UPDATE_ITEM       = 'event_update_item'
    
    def __init__(self, items=None, item_class=None):
       
        
        if items is None:
            items = []
        
        self.items = items
        
        MutableSequence.__init__(self)
        Observable.__init__(self, log_level=LOG_LEVEL)
        
        # Each container has a item_clas property which holds the
        # class of the type of ModelItem that is gathered in the sequence.
        # self.item_class() will return a new instance of the exact ModelItem
        # class.
        
        self.item_class = item_class

        for item in items:
            self.connect_to_item(item)
    

    # nice commandline output for easy debugging    
    def __str__(self):
        return yaml.dump(list(self))
    
    def __repr__(self):
        return self.__str__()
        
    # implementation of necessary abstract methods of MutableSequence
    # ---------------------------------------------------------------
    def __len__(self):
        return len(self.items)
                
    def __getitem__(self, index):
        return self.items[index]

    def __setitem__(self, index, item):
        # if item in self: # all items must be unique
        #     raise IndexError()
        if not isinstance(item, self.item_class):
            raise TypeError(f'Item must be type {self.item_class.__name__}')
        
        self[index].disconnect(id(self))
        self.items[index] = item        
        self.connect_to_item(self[index])
    
    def __delitem__(self, index):        
        item = self[index]
        self.emit(self.EVENT_WILL_REMOVE_ITEM, item)
        
        item.disconnect(id(self))
        
        del self.items[index]

        self.emit(self.EVENT_REMOVE_ITEM, (index, item))
    
    def insert(self, index, item):
        # if item in self: # all items must be unique
        #     raise IndexError()
        if not isinstance(item, self.item_class):
            raise TypeError(f'Item must be type {self.item_class.__name__}')
            
        self.items.insert(index, item)
        self.connect_to_item(item)
        self.emit(self.EVENT_ADD_ITEM, item)
    #-----------------------------------------------------------------

    # methods for handling events
    def emit_update(self, event_data):
        # EVENT_UPDATE_ITEM is always emitted when a model item in the
        # sequence emits 'event_update'
        self.emit(self.EVENT_UPDATE_ITEM, event_data=event_data)
    
    def connect_to_item(self, item):
        # Connect to events of a ModelItem
        
        # mutable sequence is not hashable and cannot be used as a key in 
        # dictionary that maps the callbacks. Therefore use id which is unique
        item.connect(id(self), item.EVENT_UPDATE, self.emit_update)
        item.connect(id(self), item.EVENT_DELETE, self.delete_callback)
        
    def delete_callback(self, model_item):
        # A model item can be deleted by the del command. If a model item
        # will be deleted it will also be removed from the sequence
        if model_item not in self:
            pass
            # already taken care of
        else:
            # remove item from self and generate events for gui
            self.remove(model_item)

    def add_new_item(self, **kwargs): 
        # used by gui
        new_item = self.item_class(**kwargs)
        
        # note append will trigger insert and the insert method takes
        # care of connectint to the new model item
        self.append(new_item)
        
        return new_item
    
class NamedModelItemSequence(ModelItemSequence):    
    # Sequence for SourceNM, SourceXray, SourceCT, CriticalPoint
    #
    # The name attribute is somewhat special. The name attribute should be
    # unique for all items in the sequence. This object will call the
    # force_unique_name method on a item each time a ModelItem is added tot the 
    # sequence with a duplicate name
    
    def __setitem__(self, index, item):
        if not isinstance(item, self.item_class):
            raise TypeError(f'Item must be type {self.item_class.__name__}')
        item = self.force_unique_name(item)
        super().__setitem__(index, item)
    
    def insert(self, index, item):
        self.logger.debug(f'Adding item {item}')
        if not isinstance(item, self.item_class):
            raise TypeError(f'Item must be type {self.item_class.__name__}')
        self.force_unique_name(item)
        super().insert(index, item)

    @property
    def names(self):
        # convenient for gui to get all names easily
        return [item.name for item in self]
    
    def emit_update(self, event_data):
        # If a model item changes it names. Catch the name change here
        # and rename to a new unique name. 
        
        model_item, label, old_value, value = event_data
        
        if label == labels.NAME:
            names = [item.name for item in self]
            if names.count(value) > 1: # model item was assigned duplicate name
                self.force_unique_name(model_item)
                return  # renaming will trigger new update
            
        super().emit_update(event_data)
                
    def get_item_by_name(self, name):
        # convenient for gui to get item easily
        if name not in self.names:
            msg = f'No ModelItem exists with name: {name}'
        if self.names.count(name) > 1:
            # should not be possible, debugging only
            msg = f'Multiple ModelItems exists with name: {name}'
            raise KeyError(msg)
        return self[self.names.index(name)]
        
    def get_new_name(self):
        # Just add a counter to the defualt name for the new name. If new name 
        # already in self.names then the counter is increased until the
        # new name is not already in self.names
        new_name =  self.item_class.default_name + ' 1'
        i = 1
        while new_name in [item.name for item in self]:
            i += 1
            new_name = self.item_class.default_name + ' ' + str(i)
        return new_name     
    
    def force_unique_name(self, model_item):
        # rename model_item if model_item.name already in model_item.names
        for item in self:
            # if item is model_item:
            #     continue
            if model_item.name == item.name:
                model_item.name = self.get_new_name()
    
        return model_item
    
class DefaultNamedModelItemSequence(NamedModelItemSequence):
    # shieldings and clearance must always have an 'empty' or 
    # default item. THis object takes care that the item at index 0 cannot
    # be deleted. Deletion should be prevented by gui.
    
    empty_item_label = 'DefaultEmptyItemLabel'
    
    def __init__(self, items=None, item_class=None):
        super().__init__(items=items, item_class=item_class)
        
        # Always set the empty item at index zero
        if self.empty_item_label not in self.names:
           self.insert(0, self.get_empty_item())
        elif self.names.index(self.empty_item_label) != 0:
            # probably wont happen
            
            msg = 'ModelItem {self.empty_item_label} should be at index 0'
            raise IndexError(msg)
        
    def get_empty_item(self):
        return self.item_class(name=self.empty_item_label)
        
    def __delitem__(self, index):
          if index == 0:
              # Should be prevented by gui, raise error for debugging
              raise IndexError()
          else:
              super().__delitem__(index)

class MaterialItemSequence(DefaultNamedModelItemSequence):
    empty_item_label = labels.EMPTY_MATERIAL
    
    
class ShieldingItemSequence(DefaultNamedModelItemSequence):
    empty_item_label = labels.EMPTY_SHIELDING

    # get_shielding by name can be renamed to get_item_by_name in all code
    # keep for now
    def get_shielding_by_name(self, shielding_name):
        return super().get_item_by_name(shielding_name)
    
    @property
    def used_materials(self):
        # return all material names that are in this sequence
        materials = [labels.EMPTY_MATERIAL]
        for shielding in self:
            if len(shielding.materials) > 0:
                materials += [shielding.materials[0][0]]
            if len(shielding.materials) > 1:
                materials += [shielding.materials[1][0]]
        return list(set(materials))
    
    

class ClearanceItemSequence(DefaultNamedModelItemSequence):
    empty_item_label = labels.EMPTY_CLEARANCE
    
    # Legacy 
    # Ugly code to be compatible with old psp files
    @staticmethod
    def from_dict(project_dct):
        if labels.CLEARANCE in project_dct.keys():
            # Read new  clearance model
            lst = project_dct[labels.CLEARANCE]
            items = [Clearance.from_dict(item)\
                     for item in lst]
            
        else:
            # Read old biological half_life and convert to 1 fraction 
            # clearance model
            items = []
            for source in project_dct[labels.SOURCES_NM]:
                apply = source.pop(labels.APPLY_BIOLOGICAL_DECAY, False)
                half_life = source.pop(labels.BIOLOGICAL_HALFLIFE, False)
                
                if apply:
                    item = Clearance(apply_fraction1=apply,
                                                 half_life1=half_life,
                                                 fraction1=1)
                
                    if not(any([item.is_equal(other) for other in items])):
                        items += [item]
                    
                    source[labels.CLEARANCE] = item.name
                else:
                    source[labels.CLEARANCE] = labels.EMPTY_CLEARANCE
            
        return items


from pyrateshield import labels
from pyrateshield import model_items

import copy
import numpy as np



    

class Geometry(model_items.ModelItem):
    _origin_cm = None
    _pixel_size_cm = None
    _attr_dct = {
        "pixel_size_cm":     labels.PIXEL_SIZE_CM,
        "origin_cm":         labels.ORIGIN_CM,
        'locked':            labels.LOCKED}
    
    _attr_defaults = {
        labels.PIXEL_SIZE_CM: 1,
        labels.ORIGIN_CM: model_items.ZERO_POSITION,
        labels.LOCKED: False}

    
    def get_extent(self, image):
        if image is None or self.pixel_size_cm is None\
            or self.origin_cm is None:
                return (0, 1, 0, 1)
        else:
            extent = (-self.origin_cm[0],
                      image.shape[1] * self.pixel_size_cm - self.origin_cm[0],
                      -self.origin_cm[1],
                      image.shape[0] * self.pixel_size_cm - self.origin_cm[1])
        return extent
    
    @property
    def vertices(self):
        return ((0, 0), (0, 0))
    
    def pixels_to_cm(self, point):
        oo = self.origin_cm
        pp = self.pixel_size_cm
        return [point[0] * pp - oo[0], point[1] * pp - oo[1]]

    def cm_to_pixels(self, point):
        oo = self.origin_cm
        pp = self.pixel_size_cm
        return [(point[0] + oo[0])/ pp, (point[1] + oo[1]) / pp]
    
    
    def to_dict(self):
        # disable origin 
        dct = super().to_dict()        
        dct.pop(labels.ORIGIN_CM, None)
        return dct
        
        
class MeasuredGeometry(Geometry):
    _distance_cm = None
    
    _attr_dct = {
        
        "origin_cm":         labels.ORIGIN_CM,
        "vertices_pixels":   labels.VERTICES_PIXELS,
        "distance_cm":       labels.REAL_WORLD_DISTANCE_CM,
        'locked':            labels.LOCKED}
    
    _attr_defaults = {
        labels.REAL_WORLD_DISTANCE_CM: 1,
        labels.ORIGIN_CM: [0, 0],
        labels.VERTICES_PIXELS: [[0, 0], [1, 0]]}
    

    
    @property
    def vertices(self):
        vv = self.vertices_pixels
        return [self.pixels_to_cm(vv[0]), self.pixels_to_cm(vv[1])]
    

    
    @property
    def distance_pixels(self):
        vvp = self.vertices_pixels
        dpixels = np.sqrt((vvp[0][0]-vvp[1][0])**2\
                          + (vvp[0][1] - vvp[1][1])**2)

        dpixels =  float(dpixels) if dpixels > 0 else 1
        return dpixels
    
    @property
    def pixel_size_cm(self):
        return self.distance_cm / self.distance_pixels
  




class DosemapStyle(model_items.ModelItem):
    _attr_dct = {
       "cmap_name": labels.CMAP_NAME,
       "vmin": labels.CMAP_MIN,
       "vmax": labels.CMAP_MAX,
       "alpha": labels.CMAP_ALPHA,
       "alpha_gradient": labels.CMAP_ALPHA_GRADIENT,
       "contour_lines": labels.CONTOUR_LINES,
    }
    _attr_defaults = {
                  labels.CMAP_NAME: "jet",
                  labels.CMAP_MIN: 0.01,
                  labels.CMAP_MAX: 10,
                  labels.CMAP_ALPHA: 0.5,
                  labels.CMAP_ALPHA_GRADIENT: True,
                  labels.CONTOUR_LINES: [
                        [0.1, "black", "dotted", 1.5, True],
                        [0.3, "black", "dashed", 1.5, True],
                        [1.0, "darkred", "solid", 1.5, True],
                        [3.0, "black", "dashdot", 1.5, True],
                        [10., "white", "dotted", 1.5, False],
                  ]
    }
    
    @classmethod
    def from_dict(cls, dct):
        ### Fix for backwards compatibility: add the is_active flag to the end
        ### of the contour_line list if it wasn't already present.
        ### #TODO: remove this classmethod after next release.
        for cl in dct[labels.CONTOUR_LINES]:
            if len(cl) == 4:
                cl.append(True)
        
        return super().from_dict(dct)
            

class Dosemap(model_items.ModelItem):
    _grid = None
    _extent = None
    _grid_matrix_size = None
    _attr_dct = {
       "grid_matrix_size": labels.GRID_MATRIX_SIZE,
       "extent": labels.EXTENT,
       "engine": labels.ENGINE,
    }
    
    _attr_defaults = {labels.GRID_MATRIX_SIZE: 100,
                      labels.EXTENT: None,
                      labels.ENGINE: labels.RADTRACER} 
    
    @classmethod
    def from_dict(cls, dct):
        # HACK to be compatible with old files for now
        # engine is defined in dosemap and in Modlel ! FIX 
        dct[labels.ENGINE] = dct.pop(labels.ENGINE, labels.PYSHIELD)
        # Legacy old psp files have enabled item for all ModelItems
        dct.pop(labels.ENABLED, None)
        dct.pop('enable', None)
        dct.pop('show', None)
        return super().from_dict(dct)
                  
    def to_grid_coords(self, coords_cm):
        shape = self.shape
        x, y = coords_cm
        x0, x1, y0, y1 = self.extent
        j = (y1-y)/(y1-y0) * shape[0] - 0.5
        i = (x-x0)/(x1-x0) * shape[1] - 0.5
        return np.array((j, i))
        
    @property
    def extent(self):
        if self._extent is None:
            self._extent = [0 , self.grid_matrix_size, 0, self.grid_matrix_size]        
        return self._extent
    
    @extent.setter
    def extent(self, extent):
        self._extent = [float(ei) for ei in extent] if extent else None
        self._grid = None
        self._shape = None
        
    @property
    def grid_matrix_size(self):
        if self._grid_matrix_size is None:
            self._grid_matrix_size = self._attr_defaults[labels.GRID_MATRIX_SIZE]
        return self._grid_matrix_size
    
    @grid_matrix_size.setter
    def grid_matrix_size(self, grid_matrix_size):
        self._grid_matrix_size = grid_matrix_size
        self._shape = None
        self._grid = None
        
    @property
    def grid(self):  
        if self._grid is None:
            grid = Grid.make_grid(shape=self.shape, 
                                  extent=self.extent,
                                  grid_matrix_size=self.grid_matrix_size)
            self._grid = grid
        return self._grid
    
    @property
    def shape(self):
        if self._shape is None:
            x0, x1, y0, y1 = self.extent
            y_size = self.grid_matrix_size
            self._shape = (int(y_size), int(y_size * (x1-x0)/(y1-y0)))
        return self._shape
    

    
def empty_image():
    return np.ones((100, 100, 3))
        
        
    
class Floorplan(model_items.ModelItem):
    # EVENT_UPDATE_IMAGE     = 'event_update_image'
    # EVENT_UPDATE_GEOMETRY  = 'event_geometry_update'
   
    _attr_dct = {
        "geometry": labels.GEOMETRY,
        "image": labels.IMAGE}
    
    _attr_defaults = {labels.GEOMETRY: Geometry,
                      labels.IMAGE: empty_image}
        
    _image = None    


    @property
    def extent(self):
        return self.geometry.get_extent(self.image)


    def to_dict(self):
        dct = super().to_dict()
        dct[labels.GEOMETRY] = dct[labels.GEOMETRY].to_dict()
        return dct
    
    @classmethod
    def from_dict(cls, dct):
        if labels.PIXEL_SIZE_CM in dct[labels.GEOMETRY].keys():
            dct[labels.GEOMETRY] = Geometry.from_dict(dct[labels.GEOMETRY])
        else:
            dct[labels.GEOMETRY] = MeasuredGeometry.from_dict(dct[labels.GEOMETRY])
        return super().from_dict(dct)
    
    


class Grid:
    _distance_map = None
    def __init__(self, X, Y):
        self.X = X
        self.Y = Y
        
    def distance_map_meters(self, x=0, y=0):
        x, y = (float(x), float(y))

        if self._distance_map is None:
            # cache results in dict when re-used
            self._distance_map = {}

        if (x, y) not in self._distance_map.keys():
            # no distance map yet calculated for point x, y
            distance_map = np.sqrt( ((self.X-x)/100)**2\
                                   + ((self.Y - y) / 100)**2)
            # add distance map for x, y to cache
            self._distance_map[(x, y)] = distance_map

        # return distance map
        return self._distance_map[(x, y)]
    
    @classmethod
    def make_grid(cls, shape=None, extent=None, grid_matrix_size=None):
        x0, x1, y0, y1 = extent
        xcoords = x0 + (np.arange(shape[1])+0.5)*(x1-x0)/shape[1]
        ycoords = y1 - (np.arange(shape[0])+0.5)*(y1-y0)/shape[0]
        return cls(*np.meshgrid(xcoords, ycoords, sparse=False))
    
    @staticmethod
    def multiply(*args):
        # used to multiple an iterable of arrays. Can be used safely if
        # iterable has just one element
        if len(args) == 1:
            return args[0]
        else:
            return np.prod(args, axis=0)
        
    @staticmethod
    def make_grid_pyshield(extent=None, grid_matrix_size=None):
        # Ga ik niet meer gebruiken maar laat er nog even in staan. Als ik
        # weer tegen rare problemen aanloop kan ik het nog gebruiken
        
        # onderstaant geeft 1 op 1 hetzelfde grid nu
        
        raise DeprecationWarning()
        def get_spaced_axis(xi_range, gi):
            start, stop = xi_range
            remainder = (stop - start) % grid_matrix_size
            offset = remainder / 2
            offset = 0
            start += 0.5 * gi
            p = np.arange(start+offset, stop, step=gi)
            return p
        
        yi_range = extent[3] - extent[2]
        xi_range = extent[1] - extent[0]
        
        grid_spacing_y = yi_range / grid_matrix_size
        
        
        grid_spacing_x = xi_range / (int(xi_range / grid_spacing_y))

        xi = get_spaced_axis(extent[0:2], grid_spacing_x)

        # why is y inverted ?? (probably because y-axis increases from bottom
        # to top instead of standard top to bottom).
        yi = get_spaced_axis(extent[2:],  grid_spacing_y)[::-1]

        X, Y = np.meshgrid(xi, yi)
        
        return Grid(X, Y)
    



from pyrateshield import Logger, GLOBAL_LOG_LEVEL, model_items
from pyrateshield.gui.mpl_items_controller import MPLPoint, MPLWall
import numpy as np
#from matplotlib.backend_bases import MouseButton
LOG_LEVEL = GLOBAL_LOG_LEVEL
#LOG_LEVEL = 10


class MPLPicker(Logger):
    # Used by MPL to select objects by mouse, objects within radius are selected
    
    # 2DO dont use points but pixels
    PICKRADIUS_POINTS = 10
    
    def __init__(self, axes=None, mpl_items=None, item_selector=None):
        Logger.__init__(self, log_level=LOG_LEVEL)
        self.axes = axes
        self.mpl_items = mpl_items
        self.item_selector = item_selector
        self.enable_pick()
        
    @staticmethod
    def is_point_item(item):
        return isinstance(item, (model_items.SourceNM,
                                 model_items.SourceCT,
                                 model_items.SourceXray,
                                 model_items.CriticalPoint))
        
    def disable_pick(self):
        for item in self.mpl_items.plotted_items:
            item.disable_pick()
    
    def enable_pick(self):
        for item in self.mpl_items.plotted_items:
            item.enable_pick()
        
    @property
    def point_mpl_items(self):
        # return all enabled point items that are plotted by mpl_items
        return [item for item in self.mpl_items.plotted_items\
                if isinstance(item, MPLPoint) and item.get_pick_enabled()]
            
    @property
    def point_model_items(self):
        return [item.model_item for item in self.point_mpl_items]
    
            
    @property
    def wall_mpl_items(self):
        # return all walls that are plotted by mpl_items
        items = [item for item in self.mpl_items.plotted_items\
                 if isinstance(item, MPLWall) and item.get_pick_enabled()]
        if self.mpl_items.pixel_size.get_pick_enabled:
            items += [self.mpl_items.pixel_size]
        return items
            
    @property
    def wall_model_items(self):
        return [item.model_item for item in self.wall_mpl_items]
    
    @staticmethod
    def lineseg_dists(p, a, b):
        """Cartesian distance from point to line segment
    
        Edited to support arguments as series, from:
        https://stackoverflow.com/a/54442561/11208892
    
        Args:
            - p: np.array of single point, shape (2,) or 2D array, shape (x, 2)
            - a: np.array of shape (x, 2)
            - b: np.array of shape (x, 2)
        """
        # normalized tangent vectors
        d_ba = b - a
        num = np.hypot(d_ba[:, 0], d_ba[:, 1]).reshape(-1, 1)
        d = np.divide(d_ba, num, where=num!=0)
    
        # signed parallel distance components
        # rowwise dot products of 2D vectors
        s = np.multiply(a - p, d).sum(axis=1)
        t = np.multiply(p - b, d).sum(axis=1)
    
        # clamped parallel distance
        h = np.maximum.reduce([s, t, np.zeros(len(s))])
    
        # perpendicular distance component
        # rowwise cross products of 2D vectors  
        d_pa = p - a
        c = d_pa[:, 0] * d[:, 1] - d_pa[:, 1] * d[:, 0]
        
        distance = np.hypot(h, c)
        
        # zero length line segments --> points
        points = a[num[:, 0]==0]
        dpoints = np.sqrt((points[:, 0] - p[0]) ** 2 + (points[:, 1] - p[1]) ** 2)
        
        
        distance[num[:, 0]==0] = dpoints
    
        return distance
    
    @property
    def pick_radius_cm(self):
        T =  self.axes.transData.inverted().transform
        pick_radius_cm = (T([self.PICKRADIUS_POINTS, 0]) - T([0, 0]))[0]
        return pick_radius_cm
    
    def pick_model_item(self, position):
        self.logger.debug(f'Pick event at position {position}')
        
        selected_item = self.item_selector.selected_item
        
        
        distance = self.selected_item_distance(position)
        
        if distance is not None and distance <= self.pick_radius_cm:
            self.logger.debug(f'Selected item selected at distance {distance}')
            item = selected_item
        else:
            wall, dw = self.get_closest_wall(position, 
                                                walls=self.wall_model_items)
            
            point, dp = self.get_closest_point(position, 
                                               points=self.point_model_items)
            
            
            self.logger.debug(f'Wall at distance {dw}')
            self.logger.debug(f'Point at distance {dp}')
            
            item = None
            
            if wall is not None and point is not None:
                if dp <= dw and dp <= self.pick_radius_cm:
                    item = point
                if dw < dp and dw <= self.pick_radius_cm:
                    item = wall
            elif point is not None:
                if dp <= self.pick_radius_cm:
                    item = point
            elif wall is not None:
                if dw <= self.pick_radius_cm:
                    item = wall
              
        vertex_index = None
        if isinstance(item, model_items.Wall):
            _, vi, di = self.get_closest_vertex(position, walls=[item])
            self.logger.debug(f'Closest vertex at : {di}')
            if di <= self.pick_radius_cm:
                self.logger.debug(f'Vertex selected index: {vi}')
                vertex_index = vi
        return item, vertex_index
        
        
    
    @staticmethod
    def get_closest_wall(position, walls=None):            
        
        
        if walls is None or len(walls) == 0:
            return None, None
        
        p = np.asarray(position, dtype=float)
        
        vertices = np.stack([wall.vertices for wall in walls])
        
        d = MPLPicker.lineseg_dists(p, vertices[:, 0, :], vertices[:, 1, :])
        
        distance = np.min(d)
        index = np.where(d==distance)[0][0]
                
        return walls[index], distance
    
    @staticmethod
    def get_closest_vertex(position, walls=None):
        if walls is None or len(walls) == 0:
            return None, None, None
        wall, _ = MPLPicker.get_closest_wall(position, walls=walls)
        vv = np.asarray(wall.vertices)
        p = position
        vd = np.sqrt((vv[:, 0] - p[0])**2 + (vv[:, 1] - p[1])**2)
        vi = np.where(vd==np.min(vd))[0][0]
        return wall, vi, np.min(vd)
    
    @staticmethod
    def get_closest_point(position, points=None):
        if points is None or len(points) == 0:
            return None, None
        
        p = position
        pp = np.asarray([point.position for point in points])
        
        d = np.sqrt((pp[:, 0]-p[0])**2 + (pp[:, 1]-p[1])**2)
        di = np.where(d==np.min(d))[0][0]
        
        return points[di], np.min(d)
        
        
    def selected_item_distance(self, position):
        selected_item = self.item_selector.selected_item
        if isinstance(selected_item, model_items.Wall):
            _, distance = self.get_closest_wall(position, [selected_item])

        elif self.is_point_item(selected_item):
            _, distance = self.get_closest_point(position, [selected_item])
   
        else:
            distance = None
        return distance
        
    
    def pick_point_item(self, position):
        if len(self.point_mpl_items) == 0:
            self.logger.debug('No point items in view')
            return None, None
        
        point_item, distance = self.get_closest_point(position, 
                                                      self.point_model_items)
        
        if distance <= self.pick_radius_cm:
            return point_item, distance
        else:
            return None, None
            
    def pick_wall_item(self, position):                
        if len(self.wall_mpl_items) == 0:
            self.logger.debug('No walls in view')
            return None, None, None

        wall, distance = self.get_closest_wall(position, self.wall_model_items)
        
        if distance > self.pick_radius_cm:
            return None, None, None
        else:
            return self.get_closest_vertex(position, walls=[wall])
       
        
            
            
          


from setuptools import setup, find_packages

APP_NAME = "pyrateshield"

# will load the __version__ variable
# Update version number in version.py
# https://stackoverflow.com/questions/2058802/how-can-i-get-the-version-defined-in-setup-py-setuptools-in-my-package
with open('pyrateshield/version.py') as f:
    for line in f:
        if line.startswith('__version__'):
            _, _, __version__ = line.replace("'", '').split()
            break

APP_SCRIPT_NAME = "%s.py" % APP_NAME.lower()


# How to publish to pypi
# 1. Update pyrateshield/version.py
# 2. Remove build and dist folder (if present)
# 3. Create source distribution: 'python setup.py sdist'
# 4. 'pip install twine' (if not installed)
# 5. Upload to pypi: 'twine upload dist/*'

def readme():
    with open('README.md') as f:
        return f.read()


setup(name='pyrateShield',
      python_requires='>=3.8',
      version=__version__, # loaded from version.py see above
      description='Generate radiation-dose maps',
      long_description=readme(),
      keywords='pyrateShield radiation radiology nuclear medicine',
      url='https://bitbucket.org/MedPhysNL/pyrateShield',
      author='Marcel Segbers, Rob van Rooij',
      author_email='msegbers@gmail.com',
      license='GNU GPLv3',
      packages=find_packages(),

      install_requires=[
          'numpy',
          'scipy',
          'matplotlib',
          'pyqt5',
          'pynrrd',
          'pyyaml',
          'scikit-image',
          'imageio',
          'pandas',
          'psutil',
          'xlsxwriter',
          'xlrd',
          'qtawesome',
          'pyperclip'],

      entry_points={
          'console_scripts': ['pyrateshield=pyrateshield.app:main'],
      },
      include_package_data=True,
      zip_safe=False)



from PyQt5.QtWidgets import (QWidget, QCheckBox, QPushButton, QVBoxLayout, 
                             QLineEdit, QSpinBox, QGridLayout, QLabel, 
                             QComboBox, QDoubleSpinBox, QColorDialog, 
                             QRadioButton, QScrollBar)
from PyQt5 import QtCore, QtGui
import qtawesome as qta
from pyrateshield import labels, Observable, GLOBAL_LOG_LEVEL
from pyrateshield.gui import io
from pyrateshield.constants import CONSTANTS

MAX_VALUE = float('inf')
MAX_INT_VALUE = 99999999



LOG_LEVEL = GLOBAL_LOG_LEVEL
#LOG_LEVEL = 10

def safe_to_int(value):
    if value == '':
        value = 0
    else:
        value = int(round(float(value)))
    return value

def safe_to_float(value):
    if value == '':
        value = 0
    else:
        value = float(value)
    return value

def safe_set_value_to_widget(widget, value):
    # Set widget to a specified value
    # do not set when value equals current value
    # will not generate events when changed
    
    if isinstance(widget, (QSpinBox, QDoubleSpinBox)):
        if widget.value() != safe_to_float(value):
            widget.setValue(safe_to_float(value))
    elif isinstance(widget, (QLineEdit, QLabel)):
        if widget.text() != str(value):
            widget.setText(str(value))
    elif isinstance(widget, QCheckBox):
        if widget.isChecked() != value:
            widget.setChecked(value)
    elif isinstance(widget, QScrollBar):
        if widget.value() != safe_to_int(value):
            widget.setValue(safe_to_int(value))
    elif isinstance(widget, QComboBox):
        if isinstance(value, str):
            value = widget.findText(str(value))
        if widget.currentIndex != safe_to_int(value):
            widget.setCurrentIndex(safe_to_int(value))
    else:
        raise TypeError(f'Unsupported widget type {type(widget)}')

class QDoubleSpinBoxInfinite(QDoubleSpinBox):

    def __init__(self, *args, **kwargs):
        super(QDoubleSpinBox, self).__init__(*args, **kwargs)

        self.setMinimum(float('-inf'))
        self.setMaximum(float('inf'))

    def keyPressEvent(self, e: QtGui.QKeyEvent):
        if e.key() == QtCore.Qt.Key_End:
            self.setValue(self.maximum())
        elif e.key() == QtCore.Qt.Key_Home:
            self.setValue(self.minimum())
        else:
            super(QDoubleSpinBox, self).keyPressEvent(e)
            
            
            
            
            
    
        
class EditViewBase(QWidget, Observable):
    EVENT_VIEW_UPDATE = 'event_view_update'
    _layout = None
    #LABEL = 'label'
    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        Observable.__init__(self, log_level=LOG_LEVEL)
        self.create_widgets()
        self.create_layout()
        self.set_callbacks()
        self.set_stretch()
        
    def set_callbacks(self):
        pass
    
    def create_widgets(self):
        pass
    
    def create_layout(self, label):
        pass
    
    def emit_change(self, event_data=None):
        self.logger.debug(f'View changed with event data: {str(event_data)}')
        self.emit(self.EVENT_VIEW_UPDATE, event_data)
        
    
    def set_stretch(self):
        self.stretch_layout = QVBoxLayout()
        
            
        self.stretch_layout.addLayout(self.layout)
        if hasattr(self, 'explanation'):
            self.help_button = QPushButton('Help')
            callback = lambda : io.show_help(title=self.LABEL,
                                             text = self.explanation)
            self.help_button.clicked.connect(callback)        
            self.stretch_layout.addWidget(self.help_button)
            
        self.stretch_layout.addStretch(1)
        self.setLayout(self.stretch_layout)
        
    @property
    def layout(self):
        if self._layout is None:
            self._layout = QGridLayout()
        return self._layout
        
    def set_enabled(self, enabled):
        index = self.layout.count()
        for i in range(index):
            widget = self.layout.itemAt(i).widget()
            widget.setEnabled(enabled)
            
        
class EditListViewBase(EditViewBase):
    _name_text = "Name:"
    EVENT_LIST_SELECTION = 'event_list_selection'
    
    def __init__(self, parent=None):
    
        super().__init__(parent)
        
        
    def create_widgets(self):
        self.list = QComboBox(self)
        
        # self.save_button = QPushButton('Save', self)
        # self.undo_button = QPushButton('Undo', self)
        #self.new_button = QPushButton('Add', self)
        #self.delete_button = QPushButton('Delete', self)
    
        self.name_label = QLabel(self._name_text)
        self.name_input = QLineEdit()
        
        self.enabled_checkbox = QCheckBox('Enabled')
        
    def set_enabled(self, enabled):
        super().set_enabled(enabled)   
        
        self.list.setEnabled(True)

    def set_callbacks(self):
        label = labels.NAME
        callback = lambda label=label: self.emit_change(label)
        self.name_input.returnPressed.connect(callback)
        
        label = labels.ENABLED
        callback = lambda event_data, labels=label: self.emit_change(label)        
        
        self.enabled_checkbox.stateChanged.connect(callback)    
        self.list.currentIndexChanged.connect(self.list_selection)

        
    def list_selection(self, _=None):
        self.logger.debug('List Selection')
        self.emit(self.EVENT_LIST_SELECTION, event_data=self.list.currentIndex())

        
    def create_layout(self):
        row = 0
        
        self.layout.addWidget(self.list, row , 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.name_label, row, 0)
        self.layout.addWidget(self.name_input, row, 1)
      
        self.create_layout_body()
        
        row  = self.layout.rowCount() + 1

        self.layout.addWidget(self.enabled_checkbox, row, 0)

        
    def create_layout_body(self):
        pass
        
    def clear(self):
        self.name_input.clear()      
    
    def emit_change(self, label):
        self.logger.debug(f'Item {label} in view changed')
        value = self.to_dict()[label]
        super().emit_change(event_data = {label: value})
        
    def to_dict(self):
        return {labels.NAME: self.name_input.text(),
                labels.ENABLED: self.enabled_checkbox.isChecked()}
    
    def from_dict(self, dct):
        #self.disable_connection(event_name=self.EVENT_VIEW_UPDATE)
        for label, value in dct.items():
            if label == labels.NAME:
                safe_set_value_to_widget(self.name_input, value)   
                self.logger.debug(f'Setting list to {value}')
                safe_set_value_to_widget(self.list, value)
            elif label == labels.ENABLED:
                safe_set_value_to_widget(self.enabled_checkbox, value)
        #self.enable_connection(event_name=self.EVENT_VIEW_UPDATE)
        
class EditListViewPositionBase(EditListViewBase):
    _position_x_text = "X [cm]:"
    _position_y_text = "Y [cm]:"
    
    
    def set_callbacks(self):
        super().set_callbacks()
        
        label = labels.POSITION
        callback = lambda _, label=label: self.emit_change(label)
        
        self.x.valueChanged.connect(callback)
        self.y.valueChanged.connect(callback)
        

    def to_dict(self):
        dct = super().to_dict()
        dct[labels.POSITION] = self.get_position()
        return dct

    def from_dict(self, dct):
       super().from_dict(dct)       
       #self.disable_connection(event_name=self.EVENT_VIEW_UPDATE)
       for label, value in dct.items():
            if label == labels.POSITION:
                self.set_position(value)
       #self.enable_connection(event_name=self.EVENT_VIEW_UPDATE)
  
    def create_widgets(self):
        super().create_widgets()

        x = QDoubleSpinBox(decimals=1)
        x.setRange(-MAX_VALUE, MAX_VALUE)
    
        y = QDoubleSpinBox(decimals=1)
        y.setRange(-MAX_VALUE, MAX_VALUE)
        
        self.x = x
        self.y = y
        self.position_label = QLabel(labels.POSITION)
        self.position_label.setStyleSheet('font-weight: bold')
        self.position_x_label = QLabel(self._position_x_text)
        self.position_y_label = QLabel(self._position_y_text)
        
        #self.position_button = QPushButton("Set Position By Mouse")
        
    def create_layout_body(self):
        row = self.layout.rowCount() + 1
        self.layout.addWidget(self.position_label, row, 0, 1, 2)
        row += 1
        self.layout.addWidget(self.position_x_label, row, 0)
        self.layout.addWidget(self.x, row, 1)
        row += 1
        self.layout.addWidget(self.position_y_label, row, 0)
        self.layout.addWidget(self.y, row, 1)
        # row += 1
        # self.layout.addWidget(self.position_button, row, 0, 1, 2)
        
    def clear(self):
        super().clear()
        self.x.clear()
        self.y.clear()
        
    def set_position(self, coords):
        safe_set_value_to_widget(self.x, coords[0])
        safe_set_value_to_widget(self.y, coords[1])
        
    def get_position(self):
        return [self.x.value(), self.y.value()]
        

class EditMaterialsView(EditListViewBase):
    LABEL = labels.MATERIALS
    explanation =\
        ('Materials can be changed or added to a limited extend. '
         'Radtracer implements a limited set of materials. Pyshield uses '
         'attenuation and buildup tables for a limited set of '
         'materials. For some materials, like "Concrete-Barite", there is no '
         'buildup table available. For example the buildup table for "Concrete" is '
         'used for "Concrete-Barite" in pyshield. '
         '\n\n'
         'Defining or changing a material is mostly usefull to define a '
         'material that has a (slightly) different density. '
         'Both pyshield and radtracer can handle variations in density accurately.')
        
    
    def create_widgets(self):
        super().create_widgets()
        
        self.density_label = QLabel('Density [g/cm^3]')
        self.density_input = QDoubleSpinBox(decimals=3)
        self.attenuation_label = QLabel(labels.ATTENUATION_TABLE)
        self.attenuation_combo = QComboBox()
        self.attenuation_combo.addItems(CONSTANTS.base_materials)
        self.buildup_combo = QComboBox()
        self.buildup_combo.addItems(CONSTANTS.buildup_materials)
        self.buildup_label = QLabel(labels.BUILDUP_TABLE)
        self.radtracer_label = QLabel('Radtracer')
        self.pyshield_label = QLabel('Pyshield')
        self.radtracer_material_label = QLabel(labels.MATERIAL)
        self.radtracer_material_combo = QComboBox()
        self.radtracer_material_combo.addItems(CONSTANTS.base_materials)
        self.new_button = QPushButton('Add', self)
        self.delete_button = QPushButton("Delete")
        
    def from_dict(self, dct):
        super().from_dict(dct)
        
        for label, value in dct.items():
            if label == labels.DENSITY:
                safe_set_value_to_widget(self.density_input, value)
            
    
            elif label == labels.RADTRACER_MATERIAL:
                safe_set_value_to_widget(self.radtracer_material_combo, value)
                
            elif label == labels.ATTENUATION_TABLE:
                safe_set_value_to_widget(self.attenuation_combo, value)
            elif label == labels.BUILDUP_TABLE:
                safe_set_value_to_widget(self.buildup_combo, value)
    
    def to_dict(self):
        dct = super().to_dict()
        dct.pop(labels.ENABLED, None)
        
        dct[labels.DENSITY] = self.density_input.value()
        dct[labels.RADTRACER_MATERIAL] = self.radtracer_material_combo.currentText()
        
        dct[labels.ATTENUATION_TABLE] = self.attenuation_combo.currentText()
        dct[labels.BUILDUP_TABLE] = self.buildup_combo.currentText()
        return dct
        
    def set_callbacks(self):
        super().set_callbacks()
        
        label = labels.DENSITY
        callback = lambda _, label=label: self.emit_change(label)
        self.density_input.valueChanged.connect(callback)
        
        label = labels.RADTRACER_MATERIAL
        callback = lambda _, label=label: self.emit_change(label)
        self.radtracer_material_combo.currentTextChanged.connect(callback)
        
        label = labels.ATTENUATION_TABLE
        callback = lambda _, label=label: self.emit_change(label)
        self.attenuation_combo.currentTextChanged.connect(callback)
        
        label = labels.BUILDUP_TABLE
        callback = lambda _, label=label: self.emit_change(label)
        self.buildup_combo.currentTextChanged.connect(callback)
        
        
    def clear(self):
        safe_set_value_to_widget(self.density_input, 0)
        
        safe_set_value_to_widget(self.radtracer_material_combo, 
                                 CONSTANTS.base_materials[0])
        
        safe_set_value_to_widget(self.attenuation_table_combo, 
                                 CONSTANTS.base_materials[0])
        
        safe_set_value_to_widget(self.buildup_table_combo, 
                                 CONSTANTS.buildup_materials[0])
        
    
    def create_layout(self):
        super().create_layout()
        
        row = self.layout.rowCount() + 1
        
        self.layout.addWidget(self.density_label, row, 0)
        self.layout.addWidget(self.density_input, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.radtracer_label, row, 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.radtracer_material_label, row, 0)
        self.layout.addWidget(self.radtracer_material_combo, row, 1)
        
        
        row += 1
        self.layout.addWidget(self.pyshield_label, row, 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.attenuation_label, row, 0)
        self.layout.addWidget(self.attenuation_combo, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.buildup_label, row, 0)
        self.layout.addWidget(self.buildup_combo, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.new_button, row, 0)
        self.layout.addWidget(self.delete_button, row, 1)
        self.enabled_checkbox.setVisible(False)
        
    
    def set_enabled(self, enabled):
        super().set_enabled(enabled)
        self.new_button.setEnabled(True)
        
        
    
        
class EditShieldingView(EditListViewBase):
    LABEL = labels.SHIELDINGS
    _color = 'r'
    _DEFAULT_THICKNESS = 1
    _DEFAULT_LINEWIDTH = 1
    
   
    def from_dict(self, dct):
        super().from_dict(dct)
       # self.disable_connection(event_name=self.EVENT_VIEW_UPDATE)
       # self.disable_connection(event_name=self.EVENT_VIEW_UPDATE)
        
        for label, value in dct.items():
            if label == labels.MATERIALS:
                self.set_materials(value)
    
            elif label == labels.COLOR:
                if self.color != value:
                    self.color = value
                
            elif label == labels.LINEWIDTH:
                safe_set_value_to_widget(self.line_width, value)
        #self.enable_connection(event_name=self.EVENT_VIEW_UPDATE)

    
    @property
    def color(self):       
        return self._color
    
    @color.setter
    def color(self, color):
        self._color = color        
        icon = qta.icon('fa5s.circle', color=color)
        
        self.color_button.setIcon(icon)
        
        self.emit_change(labels.COLOR)
    
    
    def select_color(self):
        color = QColorDialog().getColor()
        if color.isValid():
            self.color = color.name()
            
    def to_dict(self):
        dct = super().to_dict()
        dct.pop(labels.ENABLED, None)
        
        dct[labels.LINEWIDTH]   = self.line_width.value()
        dct[labels.COLOR]       = self.color
        dct[labels.MATERIALS]   = self.get_materials()
    
        
        return dct
    
    def get_materials(self):
        material1 = self.material1_list.currentText()
        thickness1 = self.thickness1_input.value()
        
        material2 = self.material2_list.currentText()
        thickness2 =  self.thickness2_input.value()
        
        materials = [[material1, thickness1], [material2, thickness2]]
        return materials
    
    
    def set_materials(self, materials):
        if len(materials) > 0:
             material1, thickness1 = materials[0]
        else:
            material1, thickness1 = [labels.EMPTY_MATERIAL, 0]
            
        if len(materials) > 1:
            material2, thickness2 = materials[1]
        else:
            material2, thickness2 = [labels.EMPTY_MATERIAL, 0]

        safe_set_value_to_widget(self.material1_list, material1)
        safe_set_value_to_widget(self.material2_list, material2)
        safe_set_value_to_widget(self.thickness1_input, thickness1)
        safe_set_value_to_widget(self.thickness2_input, thickness2)

    
    def create_widgets(self):
        super().create_widgets()
        
        #icon = qta.icon('fa5s.circle', color='red')
        self.color_button = QPushButton("Select Color", self)
        #self.color_button.setIcon(icon)

        self.material1_label = QLabel(labels.MATERIAL + ' 1')
        self.material1_list = QComboBox()

    
        self.thickness1_label = QLabel(labels.THICKNESS)
        self.thickness1_input = QDoubleSpinBox(decimals=3)
    
        self.material2_label = QLabel(labels.MATERIAL + ' 2')
        self.material2_list = QComboBox()

        # materials = [material.name for material in CONSTANTS.materials]
        
        # self.material1_list.addItems(materials)
        # self.material2_list.addItems(materials)

        self.thickness2_label = QLabel(labels.THICKNESS)
        self.thickness2_input = QDoubleSpinBox(decimals=3)
        
        self.line_width_label = QLabel(labels.LINEWIDTH)
        self.line_width = QDoubleSpinBox()
        
        self.enabled_checkbox.setVisible(False)
        self.new_button = QPushButton('Add', self)
        self.delete_button = QPushButton("Delete")
        
        
    def set_enabled(self, enabled):
        super().set_enabled(enabled)
        self.new_button.setEnabled(True)
        

    def set_callbacks(self):
        super().set_callbacks()
        
        label = labels.MATERIALS
        callback = lambda _, label=label: self.emit_change(label)
        
        self.material1_list.currentTextChanged.connect(callback)
        self.material2_list.currentTextChanged.connect(callback)
        self.thickness1_input.valueChanged.connect(callback)
        self.thickness2_input.valueChanged.connect(callback)
        
        label = labels.LINEWIDTH
        callback = lambda _, label=label: self.emit_change(label)
        self.line_width.valueChanged.connect(callback)
        
        self.color_button.clicked.connect(self.select_color)
        
        
    def create_layout_body(self):
        
        row = self.layout.rowCount() + 1
        self.layout.addWidget(self.material1_label, row, 0)
        self.layout.addWidget(self.material1_list, row, 1)
        row += 1
        self.layout.addWidget(self.thickness1_label, row, 0)
        self.layout.addWidget(self.thickness1_input, row, 1)
        row += 1
        self.layout.addWidget(self.material2_label, row, 0)
        self.layout.addWidget(self.material2_list, row, 1)
        row += 1
        self.layout.addWidget(self.thickness2_label, row, 0)
        self.layout.addWidget(self.thickness2_input, row, 1)
        
        row += 1
        self.layout.addWidget(self.color_button, row, 0, 1, 2)
        
        row += 1
        self.layout.addWidget(self.line_width_label, row, 0)
        self.layout.addWidget(self.line_width, row, 1)
        super().create_layout_body()
        
        row += 1
        
        self.layout.addWidget(self.new_button, row, 0)
        self.layout.addWidget(self.delete_button, row, 1)
        
        
    def clear(self):
        super().clear()

        self.thickness1_input.setValue(self._DEFAULT_THICKNESS)
        self.thickness2_input.setValue(self._DEFAULT_THICKNESS)
        self.line_width.setValue(self._DEFAULT_LINEWIDTH)
        
class EditWallsView(EditViewBase):
    LABEL = labels.WALLS
    
    explanation = ("To add a Wall select Wall from the toolbar. "
                   "Click and hold the left mouse button to draw a wall")
    EVENT_SCROLL = 'event_scroll'
    start_x1, start_y1, start_x2, start_y2 = ['X1 [cm]', 'Y1 [cm]', 
                                              'X2 [cm]', 'Y2 [cm]']
    
    def create_widgets(self):
        super().create_widgets()
        self.shielding_label = QLabel("Shielding")
        self.shielding_list = QComboBox()
        
        self.position_input = {}
        self.position_label = {}
        for text in (self.start_x1, self.start_y1, 
                     self.start_x2, self.start_y2):
            self.position_input[text] = QDoubleSpinBox()
            self.position_input[text].setRange(-MAX_VALUE, MAX_VALUE)
            self.position_label[text] = QLabel(text)
            
            
        self.scroll_widget = QScrollBar(QtCore.Qt.Horizontal)
        self.scroll_widget.setPageStep(1)
        
        self.index_label = QLabel()
        
    
    def emit_change(self, label):
        self.logger.debug(f'Item {label} in view changed')
        value = self.to_dict()[label]
        super().emit_change(event_data = {label: value})
        
        
    def set_callbacks(self):
        label = labels.SHIELDING
        callback = lambda _, label=label: self.emit_change(label)
        self.shielding_list.currentTextChanged.connect(callback)
        
        
        label = labels.VERTICES
        callback = lambda _, label=label: self.emit_change(label)
        
        for text in (self.start_x1, self.start_y1, 
                     self.start_x2, self.start_y2):
            self.position_input[text].valueChanged.connect(callback)
        
        self.scroll_widget.valueChanged.connect(lambda _: self.scroll())
            
    def scroll(self):
        self.emit(self.EVENT_SCROLL, self.scroll_widget.value())
        
            
    def to_dict(self):
        dct = {}
        dct[labels.VERTICES] = self.get_vertices()
        dct[labels.SHIELDING] = self.shielding_list.currentText()

        return dct
        
    def set_index(self, index=None):
        if index is None:
            return
        
        self.scroll_widget.setValue(index)
        self.index_label.setText(f'Wall index {index}')
        
    def create_layout(self): 
        row = 0
        
        self.layout.addWidget(self.scroll_widget, row , 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.index_label, row, 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.shielding_label, row, 0)
        self.layout.addWidget(self.shielding_list, row, 1)
        
        row += 1
        
        text = self.start_x1
        self.layout.addWidget(self.position_label[text], row, 0)
        self.layout.addWidget(self.position_input[text], row, 1)
        
        row += 1
        
        text = self.start_y1
        self.layout.addWidget(self.position_label[text], row, 0)
        self.layout.addWidget(self.position_input[text], row, 1)
        
        
        row += 1
        
        text = self.start_x2
        self.layout.addWidget(self.position_label[text], row, 0)
        self.layout.addWidget(self.position_input[text], row, 1)
        
        row += 1
        
        text = self.start_y2
        self.layout.addWidget(self.position_label[text], row, 0)
        self.layout.addWidget(self.position_input[text], row, 1)
        
            
    def from_dict(self, dct):
        #self.disable_connection(event_name=self.EVENT_VIEW_UPDATE)
        for label, value in dct.items():
            if label == labels.SHIELDING:
                safe_set_value_to_widget(self.shielding_list, 
                                         dct[labels.SHIELDING])      
            elif label == labels.VERTICES:
                self.set_vertices(value)
        #self.enable_connection(event_name=self.EVENT_VIEW_UPDATE)
    
        
    def get_vertices(self):
        return [[self.position_input[self.start_x1].value(),
                 self.position_input[self.start_y1].value()],
                [self.position_input[self.start_x2].value(),
                 self.position_input[self.start_y2].value()]]
        
    def set_vertices(self, vertices):
        v1, v2 = vertices
        x1, y1 = v1
        x2, y2 = v2
        
        safe_set_value_to_widget(self.position_input[self.start_x1], x1)
        safe_set_value_to_widget(self.position_input[self.start_y1], y1)
        safe_set_value_to_widget(self.position_input[self.start_x2], x2)
        safe_set_value_to_widget(self.position_input[self.start_y2], y2)
        
        
    def clear(self):
        for text in (self.start_x1, self.start_y1, 
                     self.start_x2, self.start_y2):
            
            self.position_input[text].setValue(0)
        self.shielding_list.setCurrentIndex(0)
        
class EditClearanceView(EditListViewBase):
    LABEL = labels.CLEARANCE
    explanation = ("\nUp to two biological fraction can be defined with a"
                   " corresponding biological halflife. For each fraction "
                   "pyrateshield will calculate an effective halflife by "
                   "combining the physical halflife and biological halflife. "
                   "If (physical) decay is not checked in the Sources NM tab "
                   "physical decay will be ignored and only biological "
                   "decay will be applied.\n\n"
                   "If the fractions add up to less than 1, no biological"
                   "decay correction will be applied to the remaining fraction"
                   "\n\n"
                   "Optionally fractions can be split in time. The first "
                   "fraction will be integrated until the split time. The "
                   "second fraction and if applicable the remaining fraction "
                   "will be integrated from the split time. Integration will "
                   "always stop after source duration.")
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.toggle_fraction1()
        self.toggle_fraction2()
        self.toggle_split_time()
     
    def set_enabled(self, enabled):
        super().set_enabled(enabled)
        self.new_button.setEnabled(True)
        self.toggle_fraction1()
        self.toggle_fraction2()
        self.toggle_split_time()
        
    def from_dict(self, dct):
        super().from_dict(dct)
        
        if labels.ENABLE_SPLIT_FRACTIONS in dct.keys():
            value = dct[labels.ENABLE_SPLIT_FRACTIONS]
            safe_set_value_to_widget(self.split_time_checkbox, value)   
        
        
        for label, value in dct.items():
            if label == labels.APPLY_FRACTION1:
                safe_set_value_to_widget(self.decay_fraction1_checkbox, value)
                self.enable_fraction1()
            elif label == labels.APPLY_FRACTION2:
                safe_set_value_to_widget(self.decay_fraction2_checkbox, value)
                self.enable_fraction2()
            elif label == labels.DECAY_FRACTION1:
                safe_set_value_to_widget(self.decay_fraction1, value)
            elif label == labels.DECAY_FRACTION2:
                safe_set_value_to_widget(self.decay_fraction2, value)
            elif label == labels.HALFLIFE1:
                safe_set_value_to_widget(self.half_life_value1, value)
            elif label == labels.HALFLIFE2:
                safe_set_value_to_widget(self.half_life_value2, value)
            
            elif label == labels.SPLIT_FRACTION_TIME:
                safe_set_value_to_widget(self.split_time_input, value)
                         
            
    def to_dict(self):
        dct = super().to_dict()
        dct[labels.APPLY_FRACTION1] = self.decay_fraction1_checkbox.isChecked()
        dct[labels.APPLY_FRACTION2] = self.decay_fraction2_checkbox.isChecked()  
        dct[labels.DECAY_FRACTION1] = self.decay_fraction1.value()
        dct[labels.DECAY_FRACTION2] = self.decay_fraction2.value()
        dct[labels.ENABLE_SPLIT_FRACTIONS] = self.split_time_checkbox.isChecked()
        dct[labels.SPLIT_FRACTION_TIME] = self.split_time_input.value()
        dct[labels.HALFLIFE1] = self.half_life_value1.value()
        dct[labels.HALFLIFE2] = self.half_life_value2.value()    
        return dct
    
    
    def create_widgets(self):
        super().create_widgets()
        
        self.decay_fraction1_checkbox = QCheckBox(labels.DECAY_FRACTION1)
        self.decay_fraction1 = QDoubleSpinBox(decimals=2)
        self.decay_fraction1.setRange(0, 1)
        self.decay_fraction1.setSingleStep(0.05)
        

        self.decay_fraction2_checkbox = QCheckBox(labels.DECAY_FRACTION2)
        self.decay_fraction2 = QDoubleSpinBox(decimals=2)
        self.decay_fraction2.setRange(0, 1)
        self.decay_fraction2.setSingleStep(0.05)
        
        # Prevent halflives to become zero
        self.half_life_label1 = QLabel(labels.HALF_LIFE)
        self.half_life_value1 = QDoubleSpinBoxInfinite(decimals=2)
        self.half_life_value1.setRange(0.01, MAX_VALUE)
        
        self.half_life_label2 = QLabel(labels.HALF_LIFE)
        self.half_life_value2 = QDoubleSpinBoxInfinite(decimals=2)
        self.half_life_value2.setRange(0.01, MAX_VALUE)
        
        self.split_time_checkbox = QCheckBox(labels.ENABLE_SPLIT_FRACTIONS)
        self.split_time_label = QLabel(labels.SPLIT_FRACTION_TIME)
        self.split_time_input = QDoubleSpinBox(decimals=1)
        self.split_time_input.setRange(0, MAX_VALUE)
        
        
        self.enabled_checkbox.setVisible(False)
        self.new_button = QPushButton('Add', self)
        self.delete_button = QPushButton("Delete")
        
    def set_callbacks(self):
        super().set_callbacks()
        label = labels.HALFLIFE1
        callback = lambda _, label=label: self.emit_change(label)        
        self.half_life_value1.valueChanged.connect(callback)
        
        label = labels.HALFLIFE2
        callback = lambda _, label=label: self.emit_change(label)        
        self.half_life_value2.valueChanged.connect(callback)

        label = labels.SPLIT_FRACTION_TIME
        callback = lambda _, label=label: self.emit_change(label)        
        self.split_time_input.valueChanged.connect(callback)
        
        label = labels.DECAY_FRACTION1
        callback = lambda _, label=label: self.emit_change(label)        
        self.decay_fraction1.valueChanged.connect(callback)
        
        label = labels.DECAY_FRACTION2
        callback = lambda _, label=label: self.emit_change(label)        
        self.decay_fraction1.valueChanged.connect(callback)
        
        
        self.decay_fraction1_checkbox.toggled.connect(self.toggle_fraction1)
        self.decay_fraction2_checkbox.toggled.connect(self.toggle_fraction2)
        self.split_time_checkbox.toggled.connect(self.toggle_split_time)
        
        
        
    def create_layout_body(self):
        row = self.layout.rowCount() + 1
     
        self.layout.addWidget(self.decay_fraction1_checkbox, row, 0)
        self.layout.addWidget(self.decay_fraction1, row, 1)
        row += 1
        self.layout.addWidget(self.half_life_label1, row, 0)
        self.layout.addWidget(self.half_life_value1, row, 1)    
                
        row += 1
        
        self.layout.addWidget(self.decay_fraction2_checkbox, row, 0)
        self.layout.addWidget(self.decay_fraction2, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.half_life_label2, row, 0)
        self.layout.addWidget(self.half_life_value2, row, 1)  
        
       
        row += 1
        
        self.layout.addWidget(self.split_time_checkbox, row, 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.split_time_label, row, 0)
        self.layout.addWidget(self.split_time_input, row, 1)
   
        row += 1
        
        self.layout.addWidget(self.new_button, row, 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.delete_button, row, 0, 1, 2)
      
        
                    
    def enable_fraction1(self):
        state = self.decay_fraction1_checkbox.isChecked()
        self.decay_fraction1.setEnabled(state)
        self.half_life_label1.setEnabled(state)
        self.half_life_value1.setEnabled(state)
        # if state:
        #     self.validate_fraction1()
        
    
    def enable_fraction2(self):
        state = self.decay_fraction2_checkbox.isChecked()
        self.decay_fraction2.setEnabled(state)
        self.half_life_label2.setEnabled(state)
        self.half_life_value2.setEnabled(state)
        # if state:
        #     self.validate_fraction2()
    
    def enable_split_time(self):
        state = self.split_time_checkbox.isChecked()
        self.split_time_label.setEnabled(state)
        self.split_time_input.setEnabled(state)
        
        if state:
            self.decay_fraction1_checkbox.setChecked(state)
            self.decay_fraction2_checkbox.setChecked(state)
            self.decay_fraction1.setValue(1)
            self.decay_fraction2.setValue(1)

        self.decay_fraction1_checkbox.setEnabled(not state)
        self.decay_fraction2_checkbox.setEnabled(not state)
        
        self.decay_fraction1.setEnabled(not state)
        self.decay_fraction2.setEnabled(not state)
        
                
    def toggle_split_time(self, _=None):
        label = labels.ENABLE_SPLIT_FRACTIONS
        #self.validate_fraction1()
        self.emit_change(label)
        self.enable_split_time()
            
    def toggle_fraction1(self, _=None):
        label = labels.APPLY_FRACTION1
        #self.validate_fraction1()
        self.emit_change(label)
        self.enable_fraction1()
        
    def toggle_fraction2(self, _=None):        
        label = labels.APPLY_FRACTION2
        #self.validate_fraction2()
        self.emit_change(label)
        self.enable_fraction2()
 
        
class EditSourcesNMView(EditListViewPositionBase):
    LABEL = labels.SOURCES_NM
    explanation = ("To add a new source select Source NM from the toolbar and "
                   "click on the floorplan to add a source at that position."
                   "\n\n"
                   "When setting self shielding to 'Body', pyshield will "
                   "assume 10 cm of water as additional shielding (buildup "
                   "and attenuation). Radtracer will use a pre-simulated "
                   "spectrum after additional 10 cm of "
                   "shielding with water. Note that large differences in results "
                   "may occur between pyshield and radtracer when setting self "
                   "shielding to 'Body.'\n\n"
                   "Note that setting the self shielding to 'Body' may increase "
                   "the dose rates for pyshield for some isotopes. This is due "
                   "to the buildup in 10 cm of water. Using a user defined fixed "
                   "factor could provide a better results for pyshield.")
                   
    
    def from_dict(self, dct):
        super().from_dict(dct)
        #self.disable_connection(event_name=self.EVENT_VIEW_UPDATE)
        for label, value in dct.items():
            if label == labels.DURATION:
                safe_set_value_to_widget(self.duration, value)
            elif label == labels.ACTIVITY:
                safe_set_value_to_widget(self.activity, value)
            elif label == labels.ISOTOPE:
                safe_set_value_to_widget(self.isotope_input, value)
                halflife = CONSTANTS.get_isotope_by_name(value).half_life
                safe_set_value_to_widget(self.isotope_half_life_value, halflife)
            elif label == labels.SELF_SHIELDING:
                if isinstance(value, str):
                    safe_set_value_to_widget(self.self_shielding_list, value)
                else:
                    safe_set_value_to_widget(self.self_shielding_list, labels.SELF_SHIELDING_FACTOR)
                    safe_set_value_to_widget(self.self_shielding_factor_input, value)                
            elif label == labels.NUMBER_OF_EXAMS:
                safe_set_value_to_widget(self.number_of_exams_input, value)
            elif label == labels.APPLY_DECAY_CORRECTION:
                safe_set_value_to_widget(self.decay_correction, value)
                self.toggle_halflife_visible()
            elif label == labels.CLEARANCE:
                safe_set_value_to_widget(self.clearance_list, value)
                
    def toggle_halflife_visible(self):
        visible = self.decay_correction.isChecked()
        self.isotope_half_life_label.setEnabled(visible)
        self.isotope_half_life_value.setEnabled(visible)
        
    def set_enabled(self, enabled):
        super().set_enabled(enabled)
        
        
        if self.self_shielding_list.currentText() != labels.SELF_SHIELDING_FACTOR:
            enabled = False
        
        if enabled:
            self.toggle_halflife_visible()
       
        
        self.logger.debug(f'Setting Visble to {enabled}')
        
        
        self.set_enabled_self_shielding(enabled)
        
         
        
    def set_enabled_self_shielding(self, enabled):
        self.self_shielding_factor_input.setEnabled(enabled)
        self.self_shielding_factor_input.setVisible(enabled)
        self.self_shielding_factor_label.setEnabled(enabled)
        self.self_shielding_factor_label.setVisible(enabled)
       
                
    def to_dict(self):
        dct = super().to_dict()
        dct[labels.DURATION] = self.duration.value()
        dct[labels.ACTIVITY] =  self.activity.value()
        dct[labels.ISOTOPE] = self.isotope_input.currentText()
        
        value = self.self_shielding_list.currentText()
        if value == labels.SELF_SHIELDING_FACTOR:
            dct[labels.SELF_SHIELDING] = self.self_shielding_factor_input.value()
        else:
            dct[labels.SELF_SHIELDING] = value
            
        dct[labels.NUMBER_OF_EXAMS] = self.number_of_exams_input.value()
        dct[labels.APPLY_DECAY_CORRECTION] = self.decay_correction.isChecked()
        dct[labels.CLEARANCE] = self.clearance_list.currentText()
    
        return dct

    
    def create_widgets(self):
        super().create_widgets()
    
        self.duration = QDoubleSpinBoxInfinite()
        self.duration.setRange(0, MAX_VALUE)
        self.duration_label = QLabel(labels.DURATION)
        
        
       
        self.activity = QDoubleSpinBox(decimals=2)
        self.activity.setRange(0, MAX_VALUE)
        self.activity_label = QLabel(labels.ACTIVITY)
        
        self.isotope_input = QComboBox()
        self.isotope_input.addItems(CONSTANTS.get_isotope_name_list())
        
        self.isotope_label = QLabel(labels.ISOTOPE)
        
        self.self_shielding_list = QComboBox()
        self.self_shielding_list.addItems(CONSTANTS.self_shielding_options)
        self.self_shielding_label = QLabel(labels.SELF_SHIELDING)
        self.self_shielding_factor_label = QLabel(labels.SELF_SHIELDING_FACTOR)
        self.self_shielding_factor_input = QDoubleSpinBox()
        self.self_shielding_factor_input.setSingleStep(0.01)
        self.self_shielding_factor_input.setRange(0, MAX_VALUE)
        self.self_shielding_factor_input.setDecimals(2)
        self.self_shielding_factor_input.setValue(1.0)
        
        
        self.number_of_exams_label = QLabel(labels.NUMBER_OF_EXAMS)
        self.number_of_exams_input = QSpinBox()
        self.number_of_exams_input.setRange(0, MAX_INT_VALUE)
        

        self.decay_correction = QCheckBox(labels.APPLY_DECAY_CORRECTION)
        
        self.isotope_half_life_label = QLabel(labels.HALF_LIFE)
        self.isotope_half_life_value = QLabel("0")
        
        self.clearance_label = QLabel('Clearance Model')
        self.clearance_list = QComboBox()
        

     
    def set_callbacks(self):
        super().set_callbacks()
        
        label = labels.DURATION
        callback = lambda _, label=label: self.emit_change(label)
        self.duration.valueChanged.connect(callback)
        
        label = labels.ACTIVITY
        callback = lambda _, label=label: self.emit_change(label)
        self.activity.valueChanged.connect(callback)
        
        label = labels.ISOTOPE
        callback = lambda _, label=label: self.emit_change(label)
        self.isotope_input.currentTextChanged.connect(callback)
        
        label = labels.SELF_SHIELDING
        callback = lambda _, label=label: self.change_self_shielding(label)
        self.self_shielding_list.currentTextChanged.connect(callback)
        callback = lambda _, label=label: self.emit_change(label)
        self.self_shielding_factor_input.valueChanged.connect(callback)
        
        
        label = labels.NUMBER_OF_EXAMS
        callback = lambda _, label=label: self.emit_change(label)
        self.number_of_exams_input.valueChanged.connect(callback)
        
        label = labels.APPLY_DECAY_CORRECTION
        callback = lambda _, label=label: self.emit_change(label)        
        self.decay_correction.stateChanged.connect(callback)
        
        label = labels.CLEARANCE
        callback = lambda _, label=label: self.emit_change(label)        
        self.clearance_list.currentIndexChanged.connect(callback)
    
    def change_self_shielding(self, label):
        if self.self_shielding_list.currentText() == labels.SELF_SHIELDING_FACTOR:
            enabled = True
        else:
            enabled = False
            
        self.set_enabled_self_shielding(enabled)
        
        super().emit_change(label)
              
    def clear(self):
        super().clear()
        self.activity.clear()
        self.number_of_exams_input.clear()
        self.decay_correction.setChecked(False)
        self.isotope_half_life_value.clear()
        self.duration.clear()
        
        

    def create_layout_body(self):
        

        row = self.layout.rowCount() + 1
        
        self.layout.addWidget(self.number_of_exams_label, row, 0)
        self.layout.addWidget(self.number_of_exams_input, row, 1)
        
        row += 1
        self.layout.addWidget(self.activity_label, row, 0)
        self.layout.addWidget(self.activity, row, 1)
        row += 1
        self.layout.addWidget(self.isotope_label, row, 0)
        self.layout.addWidget(self.isotope_input, row, 1)
        row += 1
        self.layout.addWidget(self.self_shielding_label, row, 0)
        self.layout.addWidget(self.self_shielding_list, row, 1)
        row += 1
        self.layout.addWidget(self.self_shielding_factor_label, row, 0)
        self.layout.addWidget(self.self_shielding_factor_input, row, 1)
        row += 1
        self.layout.addWidget(self.duration_label, row, 0)
        self.layout.addWidget(self.duration, row, 1)
        row += 1
        self.layout.addWidget(self.decay_correction, row, 0, 1, 2)
        row += 1
        self.layout.addWidget(self.isotope_half_life_label, row, 0)
        self.layout.addWidget(self.isotope_half_life_value, row, 1)
        row += 1
        self.layout.addWidget(self.clearance_label, row, 0, 1, 2)
        row += 1
        self.layout.addWidget(self.clearance_list, row, 0, 1, 2)
       
        
       
        super().create_layout_body()


class EditCriticalPointsView(EditListViewPositionBase):
    LABEL = labels.CRITICAL_POINTS
    explanation = ("To add a new critcal point select Critical Point from "
                   "the toolbar and click on the floorplan to add a "
                   "critical point at that position.")
              
    
    def from_dict(self, dct):
        #self.disable_connection(event_name=self.EVENT_VIEW_UPDATE)
        super().from_dict(dct)
        for label, value in dct.items():
            if label == labels.OCCUPANCY_FACTOR:
                safe_set_value_to_widget(self.occupancy_factor_input, value)
        #self.enable_connection(event_name=self.EVENT_VIEW_UPDATE)
        
    def to_dict(self):
        dct = super().to_dict()
        dct[labels.OCCUPANCY_FACTOR] = self.occupancy_factor_input.value()
        return dct

    def create_widgets(self):
        super().create_widgets()
    
        self.occupancy_factor_label = QLabel("Occupancy Factor:")
        self.occupancy_factor_input = QDoubleSpinBox()
        self.occupancy_factor_input.setSingleStep(0.05)
        self.occupancy_factor_input.setRange(0, 1)
        self.occupancy_factor_input.setValue(1)
        self.clear()

    def set_callbacks(self):
        super().set_callbacks()
        label = labels.OCCUPANCY_FACTOR
        callback = lambda _, label=label: self.emit_change(label)
    
        self.occupancy_factor_input.valueChanged.connect(callback)
        
        
        
    def clear(self):
        super().clear()
        self.name_input.clear()
        self.occupancy_factor_input.clear()
        
    def create_layout_body(self):
        
        row = self.layout.rowCount() + 1

        self.layout.addWidget(self.occupancy_factor_label, row, 0)
        self.layout.addWidget(self.occupancy_factor_input, row, 1)
        super().create_layout_body() 

        
class EditSourceXrayView(EditListViewPositionBase):
    LABEL = labels.SOURCES_XRAY
    explanation = ("To add a new source select Source Xray from the toolbar and "
                   "click on the floorplan to add a source at that position.")
                   
    def from_dict(self, dct):
        #self.disable_connection(event_name=self.EVENT_VIEW_UPDATE)
        super().from_dict(dct)
        
        for label, value in dct.items():
            if label == labels.DAP:
                safe_set_value_to_widget(self.dap, value)
            elif label == labels.KVP:
                safe_set_value_to_widget(self.kvp, str(value))
            elif label == labels.NUMBER_OF_EXAMS:
                safe_set_value_to_widget(self.number_of_exams_input, value)
        #self.enable_connection(event_name=self.EVENT_VIEW_UPDATE)
        
    def to_dict(self):
        dct = super().to_dict()
        dct[labels.DAP] = self.dap.value()
        dct[labels.KVP] = safe_to_int(self.kvp.currentText())
        dct[labels.NUMBER_OF_EXAMS] = self.number_of_exams_input.value()
        return dct
    
    @property
    def kvp_labels(self):
        return [str(item.kvp) for item in CONSTANTS.xray]
    
    def create_widgets(self):
        super().create_widgets()
        self.number_of_exams_input = QSpinBox()
        self.number_of_exams_input.setRange(0, MAX_INT_VALUE)
        self.number_of_exams_label = QLabel(labels.NUMBER_OF_EXAMS)
        
        self.kvp = QComboBox()
        self.kvp.addItems(self.kvp_labels) 
        self.kvp_label = QLabel(labels.KVP)

        self.dap = QDoubleSpinBox()
        self.dap.setRange(0, MAX_VALUE)
        self.dap_label = QLabel(labels.DAP)
        
    def set_callbacks(self):
        super().set_callbacks()
        label = labels.NUMBER_OF_EXAMS
        callback = lambda _, label=label: self.emit_change(label)        
        self.number_of_exams_input.valueChanged.connect(callback)
        
        label = labels.KVP
        callback = lambda _, label=label: self.emit_change(label)        
        self.kvp.currentTextChanged.connect(callback)
        
        label = labels.DAP
        callback = lambda _, label=label: self.emit_change(label)        
        self.dap.valueChanged.connect(callback)

    def create_layout_body(self):
        
        row = self.layout.rowCount() + 1
        self.layout.addWidget(self.number_of_exams_label, row, 0)
        self.layout.addWidget(self.number_of_exams_input, row, 1)
        
        row += 1
        self.layout.addWidget(self.kvp_label, row, 0)
        self.layout.addWidget(self.kvp, row, 1)
        row += 1
        self.layout.addWidget(self.dap_label, row, 0)
        self.layout.addWidget(self.dap, row, 1)
        super().create_layout_body() 

    def clear(self):
        super().clear()

        self.number_of_exams_input.clear()
        self.kvp.setCurrentIndex(0)
        self.dap.clear()

class EditSourceCTView(EditListViewPositionBase):
    LABEL = labels.SOURCES_CT
    explanation = ("To add a new source select Source CT from the toolbar and "
                   "click on the floorplan to add a source at that position.")
    def from_dict(self, dct):
        super().from_dict(dct)
        #self.disable_connection(event_name=self.EVENT_VIEW_UPDATE)
        for label, value in dct.items():
           if label == labels.DLP:
               safe_set_value_to_widget(self.dlp, value)
           elif label == labels.KVP:
               safe_set_value_to_widget(self.kvp, str(value))
           elif label == labels.NUMBER_OF_EXAMS:
               safe_set_value_to_widget(self.number_of_exams_input, value)
           elif label == labels.CT_BODY_PART:
               safe_set_value_to_widget(self.body_part, value)
        #self.enable_connection(event_name=self.EVENT_VIEW_UPDATE)
        return dct
        
        
    
    def to_dict(self):
        dct = super().to_dict()
        dct[labels.NUMBER_OF_EXAMS] = safe_to_int(self.number_of_exams_input.value())
        dct[labels.KVP] = safe_to_int(self.kvp.currentText())
        dct[labels.CT_BODY_PART] = self.body_part.currentText()
        dct[labels.DLP] = safe_to_float(self.dlp.value())
        return dct        
    
    @property
    def kvp_labels(self):
        return [str(item.kvp) for item in CONSTANTS.ct]
    
    def create_widgets(self):
        super().create_widgets()
        self.number_of_exams_input = QSpinBox()
        self.number_of_exams_input.setRange(0, MAX_INT_VALUE)
        self.number_of_exams_label = QLabel(labels.NUMBER_OF_EXAMS)

        self.kvp = QComboBox()
        self.kvp.addItems(self.kvp_labels) 
        self.kvp_label = QLabel(labels.KVP)

        self.body_part = QComboBox()
        self.body_part.addItems(CONSTANTS.CT_body_part_options ) 
        self.body_part_label = QLabel(labels.CT_BODY_PART)

        self.dlp = QDoubleSpinBox()
        self.dlp.setRange(0, MAX_VALUE)
        self.dlp_label = QLabel(labels.DLP)

        
    def set_callbacks(self):
        super().set_callbacks()
        label = labels.NUMBER_OF_EXAMS
        callback = lambda _, label=label: self.emit_change(label)                
        self.number_of_exams_input.valueChanged.connect(callback)
        
        label = labels.KVP
        callback = lambda _, label=label: self.emit_change(label)        
        self.kvp.currentTextChanged.connect(callback)
 
        
        label = labels.DLP
        callback = lambda _, label=label: self.emit_change(label)        
        self.dlp.valueChanged.connect(callback)
                
        label = labels.CT_BODY_PART
        callback = lambda _, label=label: self.emit_change(label)        
        self.dlp.valueChanged.connect(callback)
        self.body_part.currentTextChanged.connect(callback)
        
    def create_layout_body(self):
        row = self.layout.rowCount() + 1
        self.layout.addWidget(self.number_of_exams_label, row, 0)
        self.layout.addWidget(self.number_of_exams_input, row, 1)
        row += 1
        self.layout.addWidget(self.kvp_label, row, 0)
        self.layout.addWidget(self.kvp, row, 1)
        row += 1
        self.layout.addWidget(self.body_part_label, row, 0)
        self.layout.addWidget(self.body_part, row, 1)
        row += 1
        self.layout.addWidget(self.dlp_label, row, 0)
        self.layout.addWidget(self.dlp, row, 1)
        
        super().create_layout_body()

    def clear(self):
        super().clear()
        self.number_of_exams_input.clear()
        self.kvp.setCurrentIndex(0)
        self.body_part.setCurrentIndex(0)
        self.dlp.clear()

 
class EditPixelSizeView(EditViewBase):
    LABEL = labels.PIXEL_SIZE_CM
    EVENT_GEOMETRY_RADIO_BUTTON = 'event_geometry_radio_button'
    explanation =\
("The pixel size can be set by hand or by a measurement on the floor plan " 
 "image. Measurement is done by drawing a line between two points for which "
 "the real world distance in cm is known.")


    def create_widgets(self):
        #self.explanation = QLabel(self.explanation.replace('\n', ' '))
        
        
        self.choose_fixed = QRadioButton("Set Fixed")
        
        self.choose_measured = QRadioButton("Measure")

        
        self.measure_button = QPushButton('Measure On Floorplan')
        
        self.physical_distance_label = QLabel("Real world distance [cm]")
        self.physical_distance = QDoubleSpinBox(decimals=2)
        self.physical_distance.setRange(-MAX_VALUE, MAX_VALUE)
        
        self.pixel_distance_label = QLabel("Distance [pixels]")
        self.pixel_distance = QLabel("Use Button To Measure")
        
        self.pixel_size_label = QLabel(labels.PIXEL_SIZE_CM)
        self.pixel_size = QDoubleSpinBox(decimals=2)
        self.pixel_size.setRange(-MAX_VALUE, MAX_VALUE)
        
        self.confirm_button = QPushButton("Confirm")
        
        self.choose_fixed.toggled.connect(self.radio_button)
        self.choose_measured.toggled.connect(self.radio_button)
        
        
        self.pixel_size_measured_label = QLabel(labels.PIXEL_SIZE_CM)
        self.pixel_size_measured = QLabel()
        self.set_callbacks()
        
    def set_callbacks(self):        
        label = labels.REAL_WORLD_DISTANCE_CM
        callback = lambda _, label=label: self.emit_change(label)        
        self.physical_distance.valueChanged.connect(callback)
        
       

    def radio_button(self):
        
        if self.choose_measured.isChecked():
            self.set_choose_measured()
        elif self.choose_fixed.isChecked():
            self.set_choose_fixed()
        
        event_data = {labels.LOCKED: self.choose_fixed.isChecked()}
        self.emit(self.EVENT_GEOMETRY_RADIO_BUTTON, event_data)
        
            
            
    def set_choose_fixed(self):
        self.choose_measured.setChecked(False)
        self.choose_fixed.setChecked(True)
        self.measure_button.setEnabled(False)
        self.physical_distance.setEnabled(False)
        self.pixel_size_label.setEnabled(True)
        self.pixel_size.setEnabled(True)
        self.pixel_size_measured.setText('')
        self.pixel_size_measured.setEnabled(False)
        
        
    def set_choose_measured(self):
        self.choose_measured.setChecked(True)
        self.choose_fixed.setChecked(False)
        self.measure_button.setEnabled(True)
        self.physical_distance.setEnabled(True)
        self.pixel_size.setEnabled(False)
        self.pixel_size_measured.setEnabled(True)
        self.pixel_size.setEnabled(False)
        
    def create_layout(self):
        
        row = self.layout.rowCount() + 1
        
        #self.layout.addWidget(self.explanation, row, 0, 1, 2)
        
        #row += 1
        
        self.layout.addWidget(self.choose_fixed, row, 0, 1, 2)

        row += 1
        
        self.layout.addWidget(self.pixel_size_label, row, 0)
        self.layout.addWidget(self.pixel_size, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.choose_measured, row, 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.physical_distance_label, row, 0)
        self.layout.addWidget(self.physical_distance, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.pixel_distance_label, row, 0)
        self.layout.addWidget(self.pixel_distance, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.pixel_size_measured_label, row, 0)
        self.layout.addWidget(self.pixel_size_measured, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.measure_button, row, 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.confirm_button, row, 0, 1, 2)

        
        self.set_choose_measured()
       
        
